from terrex.entity.entity import Entity
from terrex.entity.npc import NPC
from terrex.net.structure.rgb import Rgb
from terrex.net.variable.bool_var import BoolVar
from terrex.net.variable.const_var import ConstVar
from terrex.net.variable.float_var import FloatVar
from terrex.net.variable.int_var import IntVar
from terrex.net.variable.str_var import StrVar
from terrex.net.world_zone import WorldZone
from terrex.world.shape.rectangle import Rectangle


class Player(Entity):
    name: StrVar = StrVar("", max_len=20)

    # const
    DEFAULT_WIDTH: ConstVar = ConstVar(20)
    DEFAULT_HEIGHT: ConstVar = ConstVar(42)

    # size
    width: IntVar = IntVar(int(DEFAULT_WIDTH))
    height: IntVar = IntVar(int(DEFAULT_HEIGHT))

    # frame
    body_frame: Rectangle = Rectangle()
    leg_frame: Rectangle = Rectangle()

    # connection flags
    initialized: BoolVar = BoolVar(False)
    logged_in: BoolVar = BoolVar(False)

    # inventory
    inventory: list = []

    # world
    zone: WorldZone = WorldZone()

    # skin
    skin_variant: IntVar = IntVar(0)
    voice_variant: IntVar = IntVar(1)
    voice_pitch_offset: FloatVar = FloatVar(0.0)
    hair: IntVar = IntVar(0, min=0, max=228)
    hair_dye: IntVar = IntVar(255, min=0, max=255)
    accessory_visibility: IntVar = IntVar(0)
    hide_misc: BoolVar = BoolVar(False)

    # skin colors
    hair_color: Rgb = Rgb(151, 100, 69)
    skin_color: Rgb = Rgb(255, 125, 90)
    eye_color: Rgb = Rgb(105, 90, 75)
    shirt_color: Rgb = Rgb(175, 165, 140)
    under_shirt_color: Rgb = Rgb(160, 180, 215)
    pants_color: Rgb = Rgb(255, 230, 175)
    shoe_color: Rgb = Rgb(160, 105, 60)

    # difficulty flag
    difficulty: IntVar = IntVar(1)

    # biome torch flags
    using_biome_torches: BoolVar = BoolVar(False)
    happy_fun_torch_time: BoolVar = BoolVar(False)
    unlocked_biome_torches: BoolVar = BoolVar(False)
    unlocked_super_cart: BoolVar = BoolVar(False)
    enabled_super_cart: BoolVar = BoolVar(False)

    # consumables flags
    used_aegis_crystal: BoolVar = BoolVar(False)
    used_aegis_fruit: BoolVar = BoolVar(False)
    used_arcane_crystal: BoolVar = BoolVar(False)
    used_galaxy_pearl: BoolVar = BoolVar(False)
    used_gummy_worm: BoolVar = BoolVar(False)
    used_ambrosia: BoolVar = BoolVar(False)
    ate_artisan_bread: BoolVar = BoolVar(False)

    # health
    maxHP: IntVar = IntVar(100, min=0, max=500)
    currHP: IntVar = IntVar(100, min=0, max=500)

    # mana
    maxMana: IntVar = IntVar(20, min=0, max=200)
    currMana: IntVar = IntVar(20, min=0, max=400)

    # luck factors
    ladybug_luck_time_left: IntVar = IntVar(0)
    torch_luck: FloatVar = FloatVar(0.0)
    luck_potion: IntVar = IntVar(0)
    has_garden_gnome_nearby: BoolVar = BoolVar(False)
    broken_mirror_bad_luck: BoolVar = BoolVar(False)
    equipment_based_luck_bonus: FloatVar = FloatVar(0.0)
    coin_luck: FloatVar = FloatVar(0.0)
    kite_luck_level: IntVar = IntVar(0)

    # effects
    stinky: BoolVar = BoolVar(False)

    def __init__(self, name: str = "terrex"):
        self.name = name

        # frame
        self.body_frame.width = 40
        self.body_frame.height = 56
        self.leg_frame.width = 40
        self.leg_frame.height = 56

        # todo: create inventory logic
        self.inventory = []
        for _ in range(0, 72):
            self.inventory.append("Dummy Item")

    # ------------------------------
    # Player Luck Calculation
    # ------------------------------

    def get_ladybug_luck(self) -> float:
        """Calculates ladybug luck contribution (-1..1)."""
        if self.ladybug_luck_time_left > 0:
            return float(self.ladybug_luck_time_left) / NPC.lady_bug_good_luck_time
        elif self.ladybug_luck_time_left < 0:
            return -float(self.ladybug_luck_time_left) / NPC.lady_bug_bad_luck_time
        return 0.0

    def calculate_coin_luck(self) -> float:
        """Calculates stepwise coinLuck bonus."""
        cl = self.coin_luck
        if cl <= 0:
            return 0.0
        elif cl <= 0.249:
            return 0.025
        elif cl <= 2.49:
            return 0.05
        elif cl <= 24.9:
            return 0.075
        elif cl <= 249:
            return 0.1
        elif cl <= 2490:
            return 0.125
        elif cl <= 24900:
            return 0.15
        elif cl <= 249000:
            return 0.175
        else:
            return 0.2

    def calculate_total_luck(self) -> float:
        """Returns the total current luck of the player (-0.4..1.0)."""
        luck: float = 0.0
        luck += self.get_ladybug_luck() * 0.2
        luck += self.torch_luck * 0.2

        # assuming luck_potion is integer tier
        # 1 - 5 min potion
        # 2 - 10 min potion
        # 3 - 15 min potion
        luck += self.luck_potion * 0.1
        luck += self.kite_luck_level * 0.1 / 3

        if self.used_galaxy_pearl:
            luck += 0.03

        # if (LanternNight.LanternsUp)
        # {
        # 	this.luck += 0.3f;
        # }

        if self.has_garden_gnome_nearby:
            luck += 0.2

        if self.stinky:
            luck -= 0.25

        luck += self.equipment_based_luck_bonus
        luck += self.calculate_coin_luck()

        if self.broken_mirror_bad_luck:
            luck -= 0.25

        return luck

    # todo: Update (every tick), ResetEffects, UpdateBuffs



"""
AnimationID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class AnimationID(IntEnum):
    MushroomStatueTurningOn = 0
    MushroomStatueTurningOff = 1
    FakeChestOpening = 2
    VolcanosTurningOn = 3
    VolcanosTurningOff = 4




from enum import IntEnum

from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class DodgeType(IntEnum):
    NINJA = 1
    SHADOW = 2


class PlayerDodge(SyncPacket):
    id = MessageID.PlayerDodge

    def __init__(self, player_id: int = 0, ty: DodgeType = DodgeType.NINJA):
        self.player_id = player_id
        self.ty = ty

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.ty = DodgeType(reader.read_byte())

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_byte(self.ty)



from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer


@dataclass
class Sign:
    index: int
    x: int
    y: int
    text: str

    @classmethod
    def read(cls, reader: Reader) -> 'Sign':
        return cls(
            reader.read_short(),
            reader.read_short(),
            reader.read_short(),
            reader.read_dotnet_string(),
        )

    def write(self, writer: Writer) -> None:
        writer.write_short(self.index)
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_dotnet_string(self.text)



from dataclasses import dataclass, field

from terrex.net.structure.vec2 import Vec2


@dataclass
class Item:
    item_id: int = -1
    net_id: int = -1
    position: Vec2 = field(default_factory=lambda: Vec2())
    velocity: Vec2 = field(default_factory=lambda: Vec2())
    prefix: int = 0
    stacks: int = 1



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class AchievementMessageNPCKilled(ServerPacket):
    id = MessageID.AchievementMessageNPCKilled

    def __init__(self, npc_id: int = 0):
        self.npc_id = npc_id

    def read(self, reader: Reader) -> None:
        self.npc_id = reader.read_short()



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class HealOtherPlayer(SyncPacket):
    id = MessageID.HealOtherPlayer

    def __init__(self, player_id: int = 0, heal_amount: int = 0):
        self.player_id = player_id
        self.heal_amount = heal_amount

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.heal_amount = reader.read_short()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_short(self.heal_amount)



from dataclasses import dataclass

from terrex.net.ambience.sky_entity_type import SkyEntityType
from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


@dataclass()
class NetAmbienceModule(NetServerModule):
    id: int = 3
    player_id: int | None = None
    rand_next_num: int | None = None
    sky_entity_type: SkyEntityType | None = None

    @classmethod
    def create(
        cls, player_id: int, rand_next_num: int, sky_entity_type: SkyEntityType
    ) -> "NetAmbienceModule":
        obj = cls()
        obj.player_id = player_id
        obj.rand_next_num = rand_next_num
        obj.sky_entity_type = sky_entity_type
        return obj

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.rand_next_num = reader.read_int()
        self.sky_entity_type = SkyEntityType(reader.read_byte())

    def write(self, writer: Writer) -> None:
        if self.player_id is None or self.rand_next_num is None or self.sky_entity_type is None:
            raise ValueError("player_id, rand_next_num, sky_entity_type must not be None")
        writer.write_byte(self.player_id)
        writer.write_int(self.rand_next_num)
        writer.write_byte(self.sky_entity_type.value)



"""
Ported from Terraria source: Terraria.WorldBuilding.Shapes.Rectangle

Exact port with numpy meshgrid for bulk vectorized actions.
"""

import numpy as np

from .base import GenAction, GenActionBulk, GenShape, Point, RectangleArea


class Rectangle(GenShape):
    """Rectangular area shape."""

    def __init__(self, area: RectangleArea | None = None, quit_on_fail: bool = False):
        super().__init__(quit_on_fail)
        self._area = area or RectangleArea()

    @property
    def width(self) -> int:
        return self._area.width

    @width.setter
    def width(self, value: int) -> None:
        self._area.width = value

    @property
    def height(self) -> int:
        return self._area.height

    @height.setter
    def height(self, value: int) -> None:
        self._area.height = value

    @property
    def x(self) -> int:
        return self._area.left

    @property
    def y(self) -> int:
        return self._area.top

    @classmethod
    def from_size(cls, width: int, height: int, quit_on_fail: bool = False):
        """Constructor matching C#: Rectangle(int width, int height)"""
        return cls(RectangleArea(0, 0, width, height), quit_on_fail)

    def SetArea(self, area: RectangleArea):
        """SetArea as in C#"""
        self._area = area

    def perform(self, origin: Point, action: GenAction) -> bool:
        # Numpy fast path for vectorized actions
        if isinstance(action, GenActionBulk):
            xs = np.arange(origin.x + self._area.left, origin.x + self._area.right)
            ys = np.arange(origin.y + self._area.top, origin.y + self._area.bottom)
            grid_x, grid_y = np.meshgrid(xs, ys)
            return action.apply_bulk(grid_x, grid_y)

        # Exact C# logic
        for i in range(origin.x + self._area.left, origin.x + self._area.right):
            for j in range(origin.y + self._area.top, origin.y + self._area.bottom):
                if not self.unit_apply(action, origin, i, j) and self._quitOnFail:
                    return False
        return True



from terrex.id import MessageID, TileChangeType, TileID
from terrex.net.bits_byte import BitsByte
from terrex.net.enum.mode import NetMode
from terrex.net.streamer import Reader, Writer
from terrex.net.structure.tile import Tile
from terrex.net.tile_npc_data import TileNPCData
from terrex.net.tile_stack import TileStack
from terrex.packet.base import SyncPacket

tile_data = TileNPCData()


class AreaTileChange(SyncPacket):
    id = MessageID.AreaTileChange

    def __init__(
        self,
        tile_y: int = 0,
        tile_x: int = 0,
        height: int = 0,
        width: int = 0,
        change_type: TileChangeType = TileChangeType.NoneValue,
        tiles: TileStack | None = None,
    ):
        self.tile_y: int = tile_y
        self.tile_x: int = tile_x
        # self.tile_x = max(0, min(self.tile_x, World.max_tiles_x - self.width - 1))
        # self.tile_y = max(0, min(self.tile_y, World.max_tiles_y - self.height - 1))
        self.height: int = height
        self.width: int = width
        self.change_type: TileChangeType = change_type
        self.tiles: TileStack = tiles or TileStack()

    def read(self, reader: Reader):
        self.tile_x = reader.read_short()
        self.tile_y = reader.read_short()
        self.width = reader.read_byte()
        self.height = reader.read_byte()
        self.change_type = TileChangeType(reader.read_byte())

        self.tiles = TileStack()
        for x in range(self.tile_x, self.tile_x + self.width):
            for y in range(self.tile_y, self.tile_y + self.height):
                bits1 = BitsByte(reader.read_byte())
                bits2 = BitsByte(reader.read_byte())
                bits3 = BitsByte(reader.read_byte())

                tile_color = reader.read_byte() if bits2[2] else 0
                wall_color = reader.read_byte() if bits2[3] else 0

                tile_type = 0
                frameX = frameY = 0
                if bits1[0]:
                    tile_type = reader.read_ushort()

                    if tile_type > TileID.Count:
                        raise ValueError(f"Corrupted tile type {tile_type}")

                    if tile_data.tileFrameImportant[tile_type]:
                        frameX = reader.read_short()
                        frameY = reader.read_short()

                wall = reader.read_ushort() if bits1[2] else 0
                liquid = liquid_type = 0
                if bits1[3] and reader.net_mode == NetMode.SERVER:
                    liquid = reader.read_byte()
                    liquid_type = reader.read_byte()

                tile = Tile()
                tile.active = bits1[0]
                tile.wall = wall
                tile.liquid = liquid
                tile.liquid_type = liquid_type
                tile.wire = bits1[4]
                tile.half_brick = bits1[5]
                tile.actuator = bits1[6]
                tile.inactive = bits1[7]
                tile.wire2 = bits2[0]
                tile.wire3 = bits2[1]
                tile.wire4 = bits2[7]
                tile.slope = (int(bits2) >> 4) & 0x07
                tile.color = tile_color
                tile.wall_color = wall_color
                tile.fullbright_block = bits3[0]
                tile.fullbright_wall = bits3[1]
                tile.invisible_block = bits3[2]
                tile.invisible_wall = bits3[3]
                tile.type = tile_type
                tile.frame_x = frameX
                tile.frame_y = frameY

                self.tiles.set(x, y, tile)

    async def handle(self, world, player, evman):
        for (x, y), tile in self.tiles.items():
            if not isinstance(tile, Tile):
                continue
            world.tiles.set(x, y, tile)

    def write(self, writer: Writer):
        writer.write_short(self.tile_x)
        writer.write_short(self.tile_y)
        writer.write_byte(self.width)
        writer.write_byte(self.height)
        writer.write_byte(self.change_type)

        for x in range(self.tile_x, self.tile_x + self.width):
            for y in range(self.tile_y, self.tile_y + self.height):
                tile = self.tiles.get(x, y)
                if not tile:
                    tile = Tile()

                bits1 = BitsByte()
                bits2 = BitsByte()
                bits3 = BitsByte()

                bits1[0] = tile.active
                bits1[2] = tile.wall > 0
                bits1[3] = tile.liquid > 0 and writer.net_mode == NetMode.SERVER
                bits1[4] = tile.wire
                bits1[5] = tile.half_brick
                bits1[6] = tile.actuator
                bits1[7] = tile.inactive

                bits2[0] = tile.wire2
                bits2[1] = tile.wire3
                bits2[2] = tile.active and tile.color > 0
                bits2[3] = tile.wall > 0 and tile.wall_color > 0
                bits2[7] = tile.wire4
                bits2_value = (int(bits2) & 0x0F) | ((tile.slope & 0x07) << 4)
                bits2 = BitsByte(bits2_value)

                bits3[0] = tile.fullbright_block
                bits3[1] = tile.fullbright_wall
                bits3[2] = tile.invisible_block
                bits3[3] = tile.invisible_wall

                writer.write_byte(int(bits1))
                writer.write_byte(int(bits2))
                writer.write_byte(int(bits3))

                if bits2[2]:
                    writer.write_byte(tile.color)
                if bits2[3]:
                    writer.write_byte(tile.wall_color)

                if tile.active:
                    writer.write_ushort(tile.type)
                    if tile_data.tileFrameImportant[tile.type]:
                        writer.write_short(tile.frame_x)
                        writer.write_short(tile.frame_y)

                if tile.wall > 0:
                    writer.write_ushort(tile.wall)
                if tile.liquid > 0 and writer.net_mode == NetMode.SERVER:
                    writer.write_byte(tile.liquid)
                    writer.write_byte(tile.liquid_type)



"""
SurfaceBackgroundID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 07:02 UTC
"""

from terrex.id.set_factory import SetFactory
from enum import IntEnum, auto


class SurfaceBackgroundID(IntEnum):
    Forest1 = 0
    Corruption = 1
    Desert = 2
    Jungle = 3
    Ocean = 4
    CorruptDesert = 5
    Hallow = 6
    Snow = 7
    Crimson = 8
    Mushroom = 9
    Forest2 = 10
    Forest3 = 11
    Forest4 = 12
    HallowDesert = 13
    CrimsonDesert = 14
    Empty = 15
    Count = 16


class SurfaceBackgroundIDSets:
    class Main:
        factory = SetFactory(SurfaceBackgroundID.Count)
        IsDesertVariant = factory.create_bool_set(False, 2, 5, 13, 14)
        IsForest = factory.create_bool_set(False, 0, 10, 11, 12)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ServerPacket


class TileEntitySharing(ServerPacket):
    id = MessageID.TileEntitySharing

    def __init__(self):
        self.tile_entity_id: int = 0
        self.update_flag: bool = True
        self.tile_entity_type: int = 0
        self.x: int = 0
        self.y: int = 0

    def read(self, reader: Reader) -> None:
        self.tile_entity_id = reader.read_int()
        self.update_flag = reader.read_bool()
        if not self.update_flag:
            self.tile_entity_type = reader.read_byte()
            self.x = reader.read_short()
            self.y = reader.read_short()

    def write(self, writer: Writer) -> None:
        raise NotImplementedError(
            "Server does not send UpdateTileEntity (client-bound packet only)"
        )



"""
AchievementHelperID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class AchievementHelperID(auto):

    class Events(IntEnum):
        NightStart = 0
        DayStart = 1
        EclipseStart = 2
        EclipseEnd = 3
        BloodMoonStart = 4
        BloodMoonEnd = 5
        SmashDemonAltar = 6
        SmashShadowOrb = 7
        NPCMovedIn = 8
        StartHardmode = 9
        InvasionDefeatedGoblins = 10
        InvasionDefeatedPirates = 11
        InvasionDefeatedSnowman = 12
        InvasionDefeatedMartians = 13
        FrostMoonWave15 = 14
        PumpkinMoonWave15 = 15
        SurvivedSlimeRain = 16
        HousedAllNPCs = 17
        TruffleMovedIn = 18
        UnlockedGoldenChest = 19
        UnlockedBiomeChest = 20
        DefeatedMechaMayhem = 21
        TempleRaider = 22
        DefeatedOldOnesArmyDifficulty3 = 23
        TurnGnomeToStatue = 24
        ThrowAParty = 25
        HousedAllTownSlimes = 26
        TransmuteItem = 27
        DroneDiedInSpace = 28
        CompleteBestiary = 29
        DefeatMoonLordInFTW = 30
        DefeatMechdusa = 31
        RideACoffinCart = 32
        GetSunBurned = 33
        SurviveBoulderRain = 34
        SurviveHardcoreDeath = 35
        GraveMistake = 36
        PlayGuitar = 37
        SpotTheSunOnACoolDay = 38
        SpottedRainbowBoulder = 39
        TrainedTownNPCsForCombat = 40
        WearMoonLordSet = 41
        Terrarist = 42
        ShimmerVillager = 43
        FlyPastSpace = 44
        ResearchedManyItems = 45
        JojaCola = 46

    class Special(IntEnum):
        ConsumeHeart = 0
        ConsumeStar = 1
        ConsumeFruit = 2
        NotTheBees = 3
        HitByTrap = 4
        RainbowShoot = 5
        MinionArmy = 6
        SwingYoyo = 7
        SurviveHugeFall = 8
        Roadkill = 9
        ReachWorldBottom = 10
        ReachWorldTop = 11
        FoundBeeHive = 12
        FoundSpiderCave = 13
        FoundHell = 14
        FoundGlowshroomOverworld = 15
        PeekInGoldenChest = 16
        FlyAKiteOnAWindyDay = 17
        FoundGraveyard = 18
        GoLavaFishing = 19
        TalkToNPCAtMaxHappiness = 20
        PetThePet = 21
        FindAFairy = 22
        DeathByDeadmansChest = 23
        GainTorchGodsFavor = 24
        DrinkBottledWaterWhileDrowning = 25
        PlayOnASpecialSeed = 26
        PurifyEntireWorld = 27




from terrex.id import MessageID
from terrex.packet.base import ServerPacket


class FinishedConnectingToServer(ServerPacket):
    id = MessageID.FinishedConnectingToServer

    def read(self, reader):
        pass



from terrex.event.types import ItemDroppedEvent, ItemDropUpdateEvent
from terrex.id import MessageID
from terrex.item.item import Item
from terrex.net.streamer import Reader, Writer
from terrex.net.structure.vec2 import Vec2
from terrex.packet.base import SyncPacket


class SyncItem(SyncPacket):
    id = MessageID.SyncItem

    def __init__(
        self,
        item_id: int = 0,
        pos: Vec2 | None = None,
        vel: Vec2 | None = None,
        stack_size: int = 0,
        prefix: int = 0,
        no_delay: int = 0,
        item_net_id: int = 0,
    ):
        self.item_id = item_id
        self.pos = pos or Vec2(0.0, 0.0)
        self.vel = vel or Vec2(0.0, 0.0)
        self.stack_size = stack_size
        self.prefix = prefix
        self.no_delay = no_delay
        self.item_net_id = item_net_id

    def read(self, reader: Reader):
        self.item_id = reader.read_short()
        self.pos = Vec2.read(reader)
        self.vel = Vec2.read(reader)
        self.stack_size = reader.read_short()
        self.prefix = reader.read_byte()
        self.no_delay = reader.read_byte()
        self.item_net_id = reader.read_short()

    def write(self, writer: Writer):
        writer.write_short(self.item_id)
        self.pos.write(writer)
        self.vel.write(writer)
        writer.write_short(self.stack_size)
        writer.write_byte(self.prefix)
        writer.write_byte(self.no_delay)
        writer.write_short(self.item_net_id)

    async def handle(self, world, player, evman):
        # todo: fix Item props
        item = Item(
            self.item_id, self.item_net_id, self.pos, self.vel, self.prefix, self.stack_size
        )

        if self.item_id in world.items:
            evman.raise_event(ItemDropUpdateEvent(self, item))
        else:
            world.items[self.item_id] = item
            if self.item_id not in world.item_owner_index:
                world.item_owner_index[self.item_id] = 255
            evman.raise_event(ItemDroppedEvent(self, item))



from enum import IntEnum


class TeleportPylonType(IntEnum):
    SurfacePurity = 0
    Jungle = 1
    Hallow = 2
    Underground = 3
    Beach = 4
    Desert = 5
    Snow = 6
    GlowingMushroom = 7
    Victory = 8
    Underworld = 9
    Shimmer = 10



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class RequestWorldData(ClientPacket):
    id = MessageID.RequestWorldData

    def write(self, writer: Writer):
        pass

    def read(self, reader: Reader) -> None:
        pass



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class TeleportPlayerThroughPortal(SyncPacket):
    id = MessageID.TeleportPlayerThroughPortal

    def __init__(
        self,
        player_id: int = 0,
        portal_color_index: int = 0,
        new_pos_x: float = 0.0,
        new_pos_y: float = 0.0,
        vel_x: float = 0.0,
        vel_y: float = 0.0,
    ):
        self.player_id = player_id
        self.portal_color_index = portal_color_index
        self.new_pos_x = new_pos_x
        self.new_pos_y = new_pos_y
        self.vel_x = vel_x
        self.vel_y = vel_y

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.portal_color_index = reader.read_short()
        self.new_pos_x = reader.read_float()
        self.new_pos_y = reader.read_float()
        self.vel_x = reader.read_float()
        self.vel_y = reader.read_float()

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.player_id)
        writer.write_short(self.portal_color_index)
        writer.write_float(self.new_pos_x)
        writer.write_float(self.new_pos_y)
        writer.write_float(self.vel_x)
        writer.write_float(self.vel_y)



"""
Ported from Terraria source: Terraria.WorldBuilding.Shapes.Mound

Parabolic mound/hill shape.
"""

from typing import cast

import numpy as np

from .base import GenAction, GenActionBulk, GenShape, Point


class Mound(GenShape):
    """Parabolic mound/hill shape with given half-width base and max height."""

    def __init__(self, half_width: int, height: int, quit_on_fail: bool = False):
        super().__init__(quit_on_fail)
        self._halfWidth = half_width
        self._height = height

    def perform(self, origin: Point, action: GenAction) -> bool:
        half_w = float(self._halfWidth)
        xs_all = []
        ys_all = []

        # Fast bulk path
        use_bulk = isinstance(action, GenActionBulk) and not self._quitOnFail

        for i in range(-self._halfWidth, self._halfWidth + 1):
            # Parabolic height calculation (inverted parabola)
            num2 = int(-((self._height + 1) / (half_w * half_w)) * ((i + half_w) * (i - half_w)))
            num2 = max(0, min(self._height, num2))
            if use_bulk:
                xs = np.full(num2, origin.x + i)
                ys = np.arange(origin.y - num2, origin.y)
                xs_all.append(xs)
                ys_all.append(ys)
            else:
                for j in range(num2):
                    if (
                        not self.unit_apply(action, origin, origin.x + i, origin.y - j)
                        and self._quitOnFail
                    ):
                        return False

        if use_bulk and xs_all:
            xs = np.concatenate(xs_all)
            ys = np.concatenate(ys_all)
            return cast(bool, cast(GenActionBulk, action).apply_bulk(xs, ys))

        return True



from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


@dataclass()
class NetUnbreakableWallScanModule(NetServerModule):
    id: int = 14
    player_id: int | None = None
    inside_unbreakable_walls: bool = False

    @classmethod
    def create(
        cls, player_id: int, inside_unbreakable_walls: bool
    ) -> "NetUnbreakableWallScanModule":
        obj = cls()
        obj.player_id = player_id
        obj.inside_unbreakable_walls = inside_unbreakable_walls
        return obj

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.inside_unbreakable_walls = reader.read_bool()

    def write(self, writer: Writer) -> None:
        if self.player_id is None:
            raise ValueError("player_id must be int and not None")
        writer.write_byte(self.player_id)
        writer.write_bool(self.inside_unbreakable_walls)



from dataclasses import dataclass

from terrex.net.enum.teleport_pylon_operation import TeleportPylonOperation
from terrex.net.enum.teleport_pylon_type import TeleportPylonType
from terrex.net.streamer import Reader, Writer

from .net_module import NetSyncModule


@dataclass()
class NetTeleportPylonModule(NetSyncModule):
    id: int = 7
    operation: TeleportPylonOperation = TeleportPylonOperation.HandleTeleportRequest
    x: int = -1
    y: int = -1
    pylon_type: TeleportPylonType = TeleportPylonType.SurfacePurity

    @classmethod
    def create(
        cls, operation: TeleportPylonOperation, x: int, y: int, pylon_type: TeleportPylonType
    ) -> "NetTeleportPylonModule":
        obj = cls()
        obj.operation = operation
        obj.x = x
        obj.y = y
        obj.pylon_type = pylon_type
        return obj

    def read(self, reader: Reader) -> None:
        self.operation = TeleportPylonOperation(reader.read_byte())
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.pylon_type = TeleportPylonType(reader.read_byte())

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.operation.value)
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_byte(self.pylon_type.value)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ServerPacket


class AnglerQuest(ServerPacket):
    id = MessageID.AnglerQuest

    def __init__(self, quest: int = 0, completed: bool = False):
        self.quest = quest
        self.completed = completed

    def read(self, reader: Reader):
        self.quest = reader.read_byte()
        self.completed = reader.read_bool()

    def write(self, writer: Writer):
        writer.write_byte(self.quest)
        writer.write_bool(self.completed)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class MinionAttackTargetUpdate(ClientPacket):
    id = MessageID.MinionAttackTargetUpdate

    def __init__(self, player_id: int = 0, minion_target: int = 0):
        self.player_id = player_id
        self.minion_target = minion_target

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.minion_target = reader.read_short()

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.player_id)
        writer.write_short(self.minion_target)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class UpdatePlayerLuckFactors(SyncPacket):
    id = MessageID.UpdatePlayerLuckFactors

    # maximum luck: 1.0
    # minimum luck: -0.4
    def __init__(
        self,
        player_id: int = 0,
        ladybug_luck_time_left: int = 0,
        torch_luck: float = 0.0,
        luck_potion: int = 0,
        has_garden_gnome_nearby: bool = False,
        broken_mirror_bad_luck: bool = False,
        equipment_based_luck_bonus: float = 0.0,
        coin_luck: float = 0.0,
        kite_luck_level: int = 0,
    ):
        self.player_id = player_id
        self.ladybug_luck_time_left = ladybug_luck_time_left
        self.torch_luck = torch_luck
        self.luck_potion = luck_potion
        self.has_garden_gnome_nearby = has_garden_gnome_nearby
        self.broken_mirror_bad_luck = broken_mirror_bad_luck
        self.equipment_based_luck_bonus = equipment_based_luck_bonus
        self.coin_luck = coin_luck
        self.kite_luck_level = kite_luck_level

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.ladybug_luck_time_left = reader.read_int()
        self.torch_luck = reader.read_float()
        self.luck_potion = reader.read_byte()
        self.has_garden_gnome_nearby = reader.read_bool()
        self.broken_mirror_bad_luck = reader.read_bool()
        self.equipment_based_luck_bonus = reader.read_float()
        self.coin_luck = reader.read_float()
        self.kite_luck_level = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_int(self.ladybug_luck_time_left)
        writer.write_float(self.torch_luck)
        writer.write_byte(self.luck_potion)
        writer.write_bool(self.has_garden_gnome_nearby)
        writer.write_bool(self.broken_mirror_bad_luck)
        writer.write_float(self.equipment_based_luck_bonus)
        writer.write_float(self.coin_luck)
        writer.write_byte(self.kite_luck_level)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ServerPacket


class MassWireOperationPay(ServerPacket):
    id = MessageID.MassWireOperationPay

    def __init__(self, item_type: int = 0, quantity: int = 0, player_id: int = 0):
        self.item_type = item_type
        self.quantity = quantity
        self.player_id = player_id

    def read(self, reader: Reader):
        self.item_type = reader.read_short()
        self.quantity = reader.read_short()
        self.player_id = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_short(self.item_type)
        writer.write_short(self.quantity)
        writer.write_byte(self.player_id)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class AddNPCBuff(SyncPacket):
    id = MessageID.AddNPCBuff

    def __init__(self, npc_id: int = 0, buff: int = 0, time: int = 0):
        self.npc_id = npc_id
        self.buff = buff
        self.time = time

    def read(self, reader: Reader):
        self.npc_id = reader.read_short()
        self.buff = reader.read_ushort()
        self.time = reader.read_short()

    def write(self, writer: Writer):
        writer.write_short(self.npc_id)
        writer.write_ushort(self.buff)
        writer.write_short(self.time)



from terrex.id import MessageID
from terrex.net.streamer import Writer
from terrex.packet.base import ClientPacket


class MinionRestTargetUpdate(ClientPacket):
    id = MessageID.MinionRestTargetUpdate

    def __init__(self, player_id: int = 0, minion_attack_target: int = 0):
        self.player_id = player_id
        self.minion_attack_target = minion_attack_target

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_short(self.minion_attack_target)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import Packet


class Unknown15(Packet):
    id = MessageID.Unknown15

    def __init__(self, pkt_id: int = 0, version: str = ""):
        self.pkt_id = pkt_id
        self.version = version

    def read(self, reader: Reader) -> None:
        self.pkt_id = reader.read_short()
        self.version = reader.read_dotnet_string()

    def write(self, writer: Writer) -> None:
        writer.write_short(self.pkt_id)
        writer.write_dotnet_string(self.version)



from enum import IntEnum

from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class ChestAction(IntEnum):
    PLACE_CHEST = 0
    KILL_CHEST = 1
    PLACE_DRESSER = 2
    KILL_DRESSER = 3
    PLACE_CONTAINERS = 4
    KILL_CONTAINERS = 5

    @classmethod
    def read(cls, reader: Reader) -> 'ChestAction':
        return cls(reader.read_byte())

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.value)


class ChestUpdates(SyncPacket):
    id = MessageID.ChestUpdates

    def __init__(
        self,
        action: ChestAction = ChestAction.PLACE_CHEST,
        tile_x: int = 0,
        tile_y: int = 0,
        style: int = 0,
        chest_id_to_destroy: int = 0,
    ):
        self.action = action
        self.tile_x = tile_x
        self.tile_y = tile_y
        self.style = style
        self.chest_id_to_destroy = chest_id_to_destroy

    def read(self, reader: Reader):
        self.action = ChestAction.read(reader)
        self.tile_x = reader.read_short()
        self.tile_y = reader.read_short()
        self.style = reader.read_short()
        self.chest_id_to_destroy = reader.read_short()

    def write(self, writer: Writer):
        self.action.write(writer)
        writer.write_short(self.tile_x)
        writer.write_short(self.tile_y)
        writer.write_short(self.style)
        writer.write_short(self.chest_id_to_destroy)



"""
TeleportationStyleID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class TeleportationStyleID(IntEnum):
    TeleporterTile = 0
    RodOfDiscord = 1
    TeleportationPotion = 2
    RecallPotion = 3
    Portal = 4
    MagicConch = 5
    DebugTeleport = 6
    DemonConch = 7
    PotionOfReturn = 8
    TeleportationPylon = 9
    QueenSlimeHook = 10
    ShellphoneSpawn = 11
    ShimmerTownNPCTransform = 12
    MysticFrog = 13
    NoEffect = 14




from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class SyncWallPaintOrCoating(SyncPacket):
    id = MessageID.SyncWallPaintOrCoating

    def __init__(self, x: int = 0, y: int = 0, color: int = 0):
        self.x = x
        self.y = y
        self.color = color

    def read(self, reader: Reader):
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.color = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_byte(self.color)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class KillProjectile(SyncPacket):
    id = MessageID.KillProjectile

    def __init__(self, projectile_id: int = 0, owner: int = 0):
        self.projectile_id = projectile_id
        self.owner = owner

    def read(self, reader: Reader):
        self.projectile_id = reader.read_short()
        self.owner = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_short(self.projectile_id)
        writer.write_byte(self.owner)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class ItemRotationAndAnimation(SyncPacket):
    id = MessageID.ItemRotationAndAnimation

    def __init__(self, player_id: int = 0, item_rotation: float = 0.0, item_animation: int = 0):
        self.player_id = player_id
        self.item_rotation = item_rotation
        self.item_animation = item_animation

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.item_rotation = reader.read_float()
        self.item_animation = reader.read_short()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_float(self.item_rotation)
        writer.write_short(self.item_animation)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class SyncLoadout(ClientPacket):
    id = MessageID.SyncLoadout

    def __init__(self, player_id: int = 0, loadout_index: int = 0, accessory_visibility: int = 0):
        self.player_id = player_id
        self.loadout_index = loadout_index
        self.accessory_visibility = accessory_visibility

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.loadout_index = reader.read_byte()
        self.accessory_visibility = reader.read_ushort()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_byte(self.loadout_index)
        writer.write_ushort(self.accessory_visibility)



from abc import ABC

from terrex.net.structure.vec2 import Vec2


class Entity(ABC):
    id: int

    # todo: create position logic
    position: Vec2
    velocity: Vec2
    old_position: Vec2
    old_velocity: Vec2

    # direction
    direction: int = 1
    old_direction: int

    # size
    width: int
    height: int

    # wet
    wet: bool
    shimmer_wet: bool
    honey_wet: bool
    lava_wet: bool
    wet_count: int



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class UpdateTowerShieldStrengths(ServerPacket):
    id = MessageID.UpdateTowerShieldStrengths

    def __init__(
        self,
        solar_tower: int = 0,
        vortex_tower: int = 0,
        nebula_tower: int = 0,
        stardust_tower: int = 0,
    ):
        self.solar_tower = solar_tower
        self.vortex_tower = vortex_tower
        self.nebula_tower = nebula_tower
        self.stardust_tower = stardust_tower

    def read(self, reader: Reader):
        self.solar_tower = reader.read_ushort()
        self.vortex_tower = reader.read_ushort()
        self.nebula_tower = reader.read_ushort()
        self.stardust_tower = reader.read_ushort()



from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer

from .net_module import NetClientModule


@dataclass()
class NetCraftingResponseModule(NetClientModule):
    id: int = 11
    approved: bool = False

    @classmethod
    def create(cls, approved: bool) -> 'NetCraftingResponseModule':
        obj = cls()
        obj.approved = approved
        return obj

    def read(self, reader: Reader) -> None:
        raise NotImplementedError("NetClientModule does not implement read")

    def write(self, writer: Writer) -> None:
        writer.write_bool(self.approved)



from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer
from terrex.net.structure.vec2 import Vec2


@dataclass
class ParticleOrchestraSettings:
    position_in_world: Vec2
    movement_vector: Vec2
    unique_info_piece: int
    index_of_player_who_invoked_this: int

    @classmethod
    def read(cls, reader: Reader) -> 'ParticleOrchestraSettings':
        return cls(Vec2.read(reader), Vec2.read(reader), reader.read_int(), reader.read_byte())

    def write(self, writer: Writer) -> None:
        self.position_in_world.write(writer)
        self.movement_vector.write(writer)
        writer.write_int(self.unique_info_piece)
        writer.write_byte(self.index_of_player_who_invoked_this)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class RequestTileEntityInteraction(SyncPacket):
    id = MessageID.RequestTileEntityInteraction

    def __init__(self, tile_entity_id: int = 0, player_id: int = 0):
        self.tile_entity_id = tile_entity_id
        self.player_id = player_id

    def read(self, reader: Reader):
        self.tile_entity_id = reader.read_int()
        self.player_id = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_int(self.tile_entity_id)
        writer.write_byte(self.player_id)



"""
StatusID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class StatusID(IntEnum):
    Ok = 0
    LaterVersion = 1
    UnknownError = 2
    EmptyFile = 3
    DecryptionError = 4
    BadSectionPointer = 5
    BadFooter = 6




from terrex.event.types import WorldSectionUpdateEvent

from .base import EventTypeFilter


def UpdateTileSection() -> EventTypeFilter[WorldSectionUpdateEvent]:
    return EventTypeFilter(WorldSectionUpdateEvent)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class InstrumentSound(SyncPacket):
    id = MessageID.InstrumentSound

    def __init__(self, player_id: int = 0, note: float = 0.0):
        self.player_id = player_id
        self.note = note

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.note = reader.read_float()

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.player_id)
        writer.write_float(self.note)



# fmt: off
# ruff: noqa

# base
from .base import Packet as Packet, packet_registry as packet_registry

# packets
from .hello import Hello as Hello
from .kick import Kick as Kick
from .player_info import PlayerInfo as PlayerInfo
from .sync_player import SyncPlayer as SyncPlayer
from .sync_equipment import SyncEquipment as SyncEquipment
from .request_world_data import RequestWorldData as RequestWorldData
from .world_data import WorldData as WorldData
from .spawn_tile_data import SpawnTileData as SpawnTileData
from .status_text_size import StatusTextSize as StatusTextSize
from .tile_section import TileSection as TileSection
from .tile_frame_section import TileFrameSection as TileFrameSection
from .player_spawn import PlayerSpawn as PlayerSpawn
from .player_controls import PlayerControls as PlayerControls
from .player_active import PlayerActive as PlayerActive
from .unknown15 import Unknown15 as Unknown15 # null 15
from .player_life_mana import PlayerLifeMana as PlayerLifeMana
from .tile_manipulation import TileManipulation as TileManipulation
from .set_time import SetTime as SetTime
from .toggle_door_state import ToggleDoorState as ToggleDoorState
from .area_tile_change import AreaTileChange as AreaTileChange
from .sync_item import SyncItem as SyncItem
from .item_owner import ItemOwner as ItemOwner
from .sync_npc import SyncNPC as SyncNPC
from .unused_melee_strike import UnusedMeleeStrike as UnusedMeleeStrike
# null 25
# null 26
from .sync_projectile import SyncProjectile as SyncProjectile
from .damage_npc import DamageNPC as DamageNPC
from .kill_projectile import KillProjectile as KillProjectile
from .toggle_pvp import TogglePvp as TogglePvp
from .request_chest_open import RequestChestOpen as RequestChestOpen
from .sync_chest_item import SyncChestItem as SyncChestItem
from .sync_player_chest import SyncPlayerChest as SyncPlayerChest
from .chest_updates import ChestUpdates as ChestUpdates
from .player_heal import PlayerHeal as PlayerHeal
from .sync_player_zone import SyncPlayerZone as SyncPlayerZone
from .request_password import RequestPassword as RequestPassword
from .send_password import SendPassword as SendPassword
from .release_item_ownership import ReleaseItemOwnership as ReleaseItemOwnership
from .sync_talk_npc import SyncTalkNPC as SyncTalkNPC
from .item_rotation_and_animation import ItemRotationAndAnimation as ItemRotationAndAnimation
from .player_mana import PlayerMana as PlayerMana
from .mana_effect import ManaEffect as ManaEffect
# null 44
from .team_change import TeamChange as TeamChange
from .open_sign_request import OpenSignRequest as OpenSignRequest
from .open_sign_response import OpenSignResponse as OpenSignResponse
from .liquid_update import LiquidUpdate as LiquidUpdate
from .initial_spawn import InitialSpawn as InitialSpawn
from .player_buffs import PlayerBuffs as PlayerBuffs
from .special_fx import SpecialFX as SpecialFX
from .lock_and_unlock import LockAndUnlock as LockAndUnlock
from .add_npc_buff import AddNPCBuff as AddNPCBuff
from .npc_buffs import NPCBuffs as NPCBuffs
from .add_player_buff_pvp import AddPlayerBuffPvP as AddPlayerBuffPvP
from .unique_town_npc_info_sync_request import UniqueTownNPCInfoSyncRequest as UniqueTownNPCInfoSyncRequest
from .update_good_evil import UpdateGoodEvil as UpdateGoodEvil
from .instrument_sound import InstrumentSound as InstrumentSound
from .hit_switch import HitSwitch as HitSwitch
from .update_home_npc import UpdateHomeNPC as UpdateHomeNPC
from .spawn_boss_use_license_start_event import SpawnBossUseLicenseStartEvent as SpawnBossUseLicenseStartEvent
from .player_dodge import PlayerDodge as PlayerDodge
from .sync_tile_paint_or_coating import SyncTilePaintOrCoating as SyncTilePaintOrCoating
from .sync_wall_paint_or_coating import SyncWallPaintOrCoating as SyncWallPaintOrCoating
from .teleport_entity import TeleportEntity as TeleportEntity
from .heal_other_player import HealOtherPlayer as HealOtherPlayer
from .tshock_placeholder import TShockPlaceholder as TShockPlaceholder # null on official server, but may be used on TShock
from .client_uuid import ClientUUID as ClientUUID
from .chest_name import ChestName as ChestName
from .bug_catching import BugCatching as BugCatching
from .bug_releasing import BugReleasing as BugReleasing
from .travel_merchant_items import TravelMerchantItems as TravelMerchantItems
from .request_teleportation_by_server import RequestTeleportationByServer as RequestTeleportationByServer
from .angler_quest import AnglerQuest as AnglerQuest
from .angler_quest_finished import AnglerQuestFinished as AnglerQuestFinished
from .quests_count_sync import QuestsCountSync as QuestsCountSync
from .temporary_animation import TemporaryAnimation as TemporaryAnimation
from .invasion_progress_report import InvasionProgressReport as InvasionProgressReport
from .place_object import PlaceObject as PlaceObject
from .sync_player_chest_index import SyncPlayerChestIndex as SyncPlayerChestIndex
from .combat_text_int import CombatTextInt as CombatTextInt
from .net_modules import NetModules as NetModules
from .unused83 import Unused83 as Unused83
from .player_stealth import PlayerStealth as PlayerStealth
from .quick_stack_chests import QuickStackChests as QuickStackChests
from .tile_entity_sharing import TileEntitySharing as TileEntitySharing
from .tile_entity_placement import TileEntityPlacement as TileEntityPlacement
from .item_tweaker import ItemTweaker as ItemTweaker
from .item_frame_try_placing import ItemFrameTryPlacing as ItemFrameTryPlacing
# from .InstancedItem import InstancedItem
from .sync_emote_bubble import SyncEmoteBubble as SyncEmoteBubble
from .sync_extra_value import SyncExtraValue as SyncExtraValue
# social handshake 93
from .dev_commands import DevCommands as DevCommands
from .murder_someone_elses_portal import MurderSomeoneElsesPortal as MurderSomeoneElsesPortal
from .teleport_player_through_portal import TeleportPlayerThroughPortal as TeleportPlayerThroughPortal
from .achievement_message_npc_killed import AchievementMessageNPCKilled as AchievementMessageNPCKilled
from .achievement_message_event_happened import AchievementMessageEventHappened as AchievementMessageEventHappened
from .minion_rest_target_update import MinionRestTargetUpdate as MinionRestTargetUpdate
from .teleport_npc_through_portal import TeleportNPCThroughPortal as TeleportNPCThroughPortal
from .update_tower_shield_strengths import UpdateTowerShieldStrengths as UpdateTowerShieldStrengths
from .nebula_levelup_request import NebulaLevelupRequest as NebulaLevelupRequest
from .moonlord_horror import MoonlordHorror as MoonlordHorror
from .shop_override import ShopOverride as ShopOverride
from .gem_lock_toggle import GemLockToggle as GemLockToggle
from .poof_of_smoke import PoofOfSmoke as PoofOfSmoke
from .smart_text_message import SmartTextMessage as SmartTextMessage
from .wired_cannon_shot import WiredCannonShot as WiredCannonShot
from .mass_wire_operation import MassWireOperation as MassWireOperation
from .mass_wire_operation_pay import MassWireOperationPay as MassWireOperationPay
from .toggle_party import ToggleParty as ToggleParty
from .special_fx import SpecialFX as SpecialFX
from .crystal_invasion_start import CrystalInvasionStart as CrystalInvasionStart
from .crystal_invasion_wipe_all_the_thingsss import CrystalInvasionWipeAllTheThingsss as CrystalInvasionWipeAllTheThingsss
from .minion_attack_target_update import MinionAttackTargetUpdate as MinionAttackTargetUpdate
from .crystal_invasion_send_wait_time import CrystalInvasionSendWaitTime as CrystalInvasionSendWaitTime
from .player_hurt_v2 import PlayerHurtV2 as PlayerHurtV2
from .player_death_v2 import PlayerDeathV2 as PlayerDeathV2
from .combat_text_string import CombatTextString as CombatTextString
from .emoji import Emoji as Emoji
from .te_display_doll_data_sync import TEDisplayDollDataSync as TEDisplayDollDataSync
from .request_tile_entity_interaction import RequestTileEntityInteraction as RequestTileEntityInteraction
from .weapons_rack_try_placing import WeaponsRackTryPlacing as WeaponsRackTryPlacing
from .te_hat_rack_item_sync import TEHatRackItemSync as TEHatRackItemSync
from .sync_tile_picking import SyncTilePicking as SyncTilePicking
from .sync_revenge_marker import SyncRevengeMarker as SyncRevengeMarker
from .remove_revenge_marker import RemoveRevengeMarker as RemoveRevengeMarker
from .land_golf_ball_in_cup import LandGolfBallInCup as LandGolfBallInCup
from .finished_connecting_to_server import FinishedConnectingToServer as FinishedConnectingToServer
from .fish_out_npc import FishOutNpc as FishOutNpc
from .tamper_with_npc import TamperWithNPC as TamperWithNPC
from .play_legacy_sound import PlayLegacySound as PlayLegacySound
from .food_platter_try_placing import FoodPlatterTryPlacing as FoodPlatterTryPlacing
from .update_player_luck_factors import UpdatePlayerLuckFactors as UpdatePlayerLuckFactors
from .dead_player import DeadPlayer as DeadPlayer
from .sync_cavern_monster_type import SyncCavernMonsterType as SyncCavernMonsterType
from .request_npc_buff_removal import RequestNPCBuffRemoval as RequestNPCBuffRemoval
from .client_synced_inventory import ClientSyncedInventory as ClientSyncedInventory
from .set_counts_as_host_for_gameplay import SetCountsAsHostForGameplay as SetCountsAsHostForGameplay
# SetMiscEventValues 140


# todo: add all new packets from 1.4.5.0+
# ------
# LucyAxeMove (ClientPacket) 141
# Unknown update data for player (ClientPacket) 142
# AttemptToSkipWaitTime for player (ClientPacket) 143
# HaveDryadDoStardewAnimation for NPC (SyncPacket???) 144
# NPC ShimmerTransformToItem (ServerPacket?) 145
# NPC ShimmerTransformToNPC (ServerPacket?) 146
from .sync_loadout import SyncLoadout as SyncLoadout
# GetItemUsedLuckyCoin (ClientPacket) 148
# null 149
# SetOrRequestSpectating (SyncPacket) 150
# PickAnItemSlotToSpawnItemOn (ServerPacket) 0-400 151
# PlayerUseSoundSync (SyncPacket???) 152
# NPC ApplyEelWhipDoT (ServerPacket) 153
from .ping import Ping as Ping
# SendMaxChestItems (ServerPacket) 155
# DebugPlaceTileEntityKiteAnchorItem (ServerPacket???) (156
# UpdatePlayerTeam (SyncPacket) 157
# null 158
# Control Check Rope Usability (ClientPacket) 159
# SyncItemOnSection (ServerPacket???) 160

# fmt: on



from terrex.event.types import BaseEvent


class EventContext:
    def __init__(self, terrex, event: BaseEvent):
        from terrex.terrex import Terrex

        if not isinstance(terrex, Terrex):
            raise TypeError("terrex must be a Terrex instance")

        self.terrex = terrex
        self.world = terrex.world
        self.player = terrex.player

        self.event = event



from typing import Generic, TypeVar

from terrex.event.context import EventContext

from ..types import BaseEvent

E = TypeVar("E", bound=BaseEvent)


class EventFilter(Generic[E]):
    _event_type: type[E]

    def matches(self, ctx: EventContext) -> E | None:
        raise NotImplementedError

    def __and__(self, other: "EventFilter[E]") -> "EventFilter[E]":
        from .combinators import AndFilter

        return AndFilter(self, other)

    def __or__(self, other: "EventFilter[E]") -> "EventFilter[E]":
        from .combinators import OrFilter

        return OrFilter(self, other)


class EventTypeFilter(EventFilter[E]):
    def __init__(self, event_type: type[E]):
        self._event_type = event_type

    def matches(self, ctx: EventContext) -> E | None:
        if isinstance(ctx.event, self._event_type):
            return ctx.event
        return None


class StopPropagation(Exception):
    pass



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.net.structure.vec2 import Vec2

from .base import ClientPacket


class BugReleasing(ClientPacket):
    id = MessageID.BugReleasing

    def __init__(self) -> None:
        self.position: Vec2 = Vec2(0, 0)
        self.npc_type: int = 0
        self.style: int = 0

    def read(self, reader: Reader) -> None:
        self.position = Vec2(reader.read_int(), reader.read_int())
        self.npc_type = reader.read_short()
        self.style = reader.read_byte()

    def write(self, writer: Writer) -> None:
        writer.write_int(int(self.position.x))
        writer.write_int(int(self.position.y))
        writer.write_short(self.npc_type)
        writer.write_byte(self.style)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class SyncPlayerChest(SyncPacket):
    id = MessageID.SyncPlayerChest

    def __init__(self, chest_id: int = 0, x: int = 0, y: int = 0, name: str = ""):
        self.chest_id = chest_id
        self.x = x
        self.y = y
        self.name = name

    def read(self, reader: Reader):
        self.chest_id = reader.read_short()
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.name = reader.read_dotnet_string()

    def write(self, writer: Writer):
        writer.write_short(self.chest_id)
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_dotnet_string(self.name)



"""
PaintID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class PaintID(IntEnum):
    NoneValue = 0
    RedPaint = 1
    OrangePaint = 2
    YellowPaint = 3
    LimePaint = 4
    GreenPaint = 5
    TealPaint = 6
    CyanPaint = 7
    SkyBluePaint = 8
    BluePaint = 9
    PurplePaint = 10
    VioletPaint = 11
    PinkPaint = 12
    DeepRedPaint = 13
    DeepOrangePaint = 14
    DeepYellowPaint = 15
    DeepLimePaint = 16
    DeepGreenPaint = 17
    DeepTealPaint = 18
    DeepCyanPaint = 19
    DeepSkyBluePaint = 20
    DeepBluePaint = 21
    DeepPurplePaint = 22
    DeepVioletPaint = 23
    DeepPinkPaint = 24
    BlackPaint = 25
    WhitePaint = 26
    GrayPaint = 27
    BrownPaint = 28
    ShadowPaint = 29
    NegativePaint = 30
    IlluminantPaint = 31
    Old_IlluminantPaint = 31




from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class Hello(ClientPacket):
    """
    It is used for the first request and comparison by the server of the Terraria version with the client version.

    In case of an error, the server sends a Kick [2] packet and terminates the connection.
    """

    id = MessageID.Hello

    def __init__(self, version: int = 0):
        self.version = version

    def read(self, reader: Reader) -> None:
        version_str = reader.read_dotnet_string()
        self.version = int(version_str.replace("Terraria", ""))

    def write(self, writer: Writer):
        writer.write_dotnet_string("Terraria" + str(self.version))



from dataclasses import dataclass, field

from terrex.item.item import Item
from terrex.net.streamer import Reader, Writer


@dataclass
class Chest:
    index: int = 0
    x: int = 0
    y: int = 0
    name: str = ""
    max_items: int = 40
    bank_chest: bool = False
    items: list[Item] = field(default_factory=list)
    _items_got_set: bool = False

    @classmethod
    def create(
        cls, index: int = 0, x: int = 0, y: int = 0, bank: bool = False, max_items: int = 40
    ) -> 'Chest':
        chest = cls(index=index, x=x, y=y, bank_chest=bank, max_items=max_items)
        chest.fill_with_empty_instances()
        return chest

    @classmethod
    def read(cls, reader: Reader) -> 'Chest':
        chest = cls(
            reader.read_short(),
            reader.read_short(),
            reader.read_short(),
            reader.read_dotnet_string(),
        )
        chest.fill_with_empty_instances()
        return chest

    def write(self, writer: Writer) -> None:
        writer.write_short(self.index)
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_dotnet_string(self.name)

    def fill_with_empty_instances(self) -> None:
        self.items = [Item() for _ in range(self.max_items)]
        self._items_got_set = True



"""
ExtrasID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class ExtrasID(IntEnum):
    BrainScrambler = 0
    Raygun = 1
    LaserRuler = 2
    AntiGravityHookChain = 3
    SaucerDeathrayCap = 4
    BGMartianUFOSmall = 5
    BGMartianUFOBig = 6
    EventIconSnowLegion = 7
    EventIconFrostMoon = 8
    EventIconGoblinArmy = 9
    EventIconMartianMadness = 10
    EventIconPirateInvasion = 11
    EventIconPumpkinMoon = 12
    MoonLordRibs = 13
    MoonLordBackarm = 14
    MoonLordForearm = 15
    MoonLordCoreMoss = 16
    MoonLordEyeWhiteHand = 17
    MoonLordEyeWhiteHead = 18
    MoonLordEye = 19
    MartianProbeDiode = 20
    PhantasmalDeathrayBody = 21
    PhantasmalDeathrayEnd = 22
    MoonLeechBody = 23
    MoonLeechTail = 24
    MoonLordMouth = 25
    MoonLordHandMouth = 26
    MartianProbeScanWave = 27
    TEMPORARYChestOpen = 28
    MoonLordEyeMouth = 29
    CultistBossShadow = 30
    WebbedFloor = 31
    WebbedAir = 32
    CultistLightingArc = 33
    CultistRitual = 34
    CultistIceshard = 35
    StarWrath = 36
    PlayerStoned = 37
    BejeweledValkyrieWing = 38
    KingSlimeCrown = 39
    PirateShipFlag = 40
    PirateShipOar = 41
    PirateShipSail1 = 42
    PirateShipSail2 = 43
    PirateShipSail3 = 44
    PirateShipSail4 = 45
    StardustJellyfishSmall = 46
    StardustTowerMark = 47
    EmoteBubble = 48
    ScreenObfuscation = 49
    VortexBlack = 50
    MagicAura = 51
    GuideFrontal = 52
    TravellingMerchantFrontal = 53
    EquipIcons = 54
    MeteorHeadFlame = 55
    MartianWalkerCannon = 56
    AncientLight = 57
    DefenseShield = 58
    PortalGateHalo = 59
    PortalGateHalo2 = 60
    SolarBlaze = 61
    SolarBlaze2 = 62
    SolarBlaze3 = 63
    NebulaArcanumItemHold = 64
    PhantasmMuzzle = 65
    SuspiciousTentacle = 66
    Yoraiz0rDarkness = 67
    LaserGrid2 = 68
    PartyBalloonsSmall = 69
    PartyBalloonsBig = 70
    PartyBalloonsTriplet = 71
    TownNPCHats = 72
    DjinnsCurse = 73
    ForbiddenSign = 74
    BlackBolt = 75
    Tumbleweed = 76
    ElfEars = 77
    DD2ElderEye = 78
    EventIconOldOnesArmy = 79
    DD2SkeletonFlame = 80
    DD2BetsyWingFront = 81
    DD2BetsyWingBack = 82
    DD2BallistraTowerT1 = 83
    DD2BallistraTowerT2 = 84
    DD2BallistraTowerT3 = 85
    DD2LightningAuraT1 = 86
    DD2LightningAuraT2 = 87
    DD2LightningAuraT3 = 88
    ThePerfectGlow = 89
    ArkhalisMask = 90
    FallingStar = 91
    LawnMower = 92
    VoidLensOutline = 93
    MoneyTroughOutline = 94
    AmberHookChain = 95
    GolfCartBody = 96
    GolfCartWheels = 97
    SharpTears = 98
    DripplerFlail = 99
    BirdNestHead = 100
    TombstoneGlow = 101
    PortableStool = 102
    KiteRibbons = 103
    KiteRibbonsWyvern = 104
    FloatingTube = 105
    GolemLights4 = 106
    GolemLights5 = 107
    BeeMinecart = 108
    LadybugMinecart = 109
    PigronMinecart = 110
    SunflowerMinecart = 111
    HellMinecart = 112
    WitchBroom = 113
    WitchBroomTrinket = 114
    ShroomMinecart = 115
    ShroomMinecartGlow = 116
    AmethystMinecart = 117
    TopazMinecart = 118
    SapphireMinecart = 119
    EmeraldMinecart = 120
    RubyMinecart = 121
    DiamondMinecart = 122
    AmberMinecart = 123
    BeetleMinecart = 124
    MeowmereMinecart = 125
    PartyMinecart = 126
    PirateMinecart = 127
    SteampunkMinecart = 128
    BloodNautilusMouth = 129
    MapBiomeIcons = 130
    BloodNautilusCallForHelp = 131
    KiteRibbonsBoneSerpent = 132
    KiteRibbonsWorldFeeder = 133
    LanternSkyLanterns = 134
    KiteRibbonsManEater = 135
    KiteRibbonsJellyfishBlue = 136
    KiteRibbonsJellyfishPink = 137
    KiteRibbonsShark = 138
    KiteRibbonsSandShark = 139
    KiteRibbonsGoldfish = 140
    PartyCartBalloons = 141
    PirateMinecartFlag = 142
    PlayerUnicornHorn = 143
    KiteRibbonsAngryTrapper = 144
    KiteRibbonsKoi = 145
    KiteRibbonsCrawltipede = 146
    KiteRibbonsSpectrum = 147
    KiteRibbonsEyeball = 148
    FlamingoMount = 149
    CoffinMinecart = 150
    CoffinMinecartLid = 151
    SafemanSun = 152
    JimCloak = 153
    SquirrelHookChain = 154
    DiggingMoleMinecart = 155
    HallowBossGradient = 156
    HallowBossWings = 157
    HallowBossArmsLeft = 158
    HallowBossWingsBack = 159
    HallowBossArmsRight = 160
    PaintedHorseMount = 161
    MajesticHorseMount = 162
    DarkHorseMount = 163
    PogoStickMount = 164
    DiggingMoleMinecartArm = 165
    PirateShipMount = 166
    SpookyWoodMount = 167
    SantankMount = 168
    WallOfFleshGoatMount = 169
    DarkMageBookMount = 170
    PlayerRainbowWings = 171
    LavaSharkMount = 172
    PotionOfReturnMinimapIcon = 173
    KeybrandRing = 174
    PotionOfReturnMinimapHome = 175
    GolfBallMinimapOutline = 176
    QueenSlimeCrown = 177
    FairyQueenLance = 178
    Voronoi = 179
    QueenSlimeGradient = 180
    PylonCrystals = 181
    PylonMapIcons = 182
    PotionOfReturnGateIn = 183
    PotionOfReturnGateOut = 184
    QueenSlimeWing = 185
    QueenSlimeCrystalCore = 186
    HallowBossTentacles = 187
    HallowBossSkirt = 188
    FlameLashTrailShape = 189
    FlameLashTrailErosion = 190
    FlameLashTrailColorGradient = 191
    MagicMissileTrailColorGradient = 192
    MagicMissileTrailErosion = 193
    MagicMissileTrailShape = 194
    RainbowRodTrailColorGradient = 195
    RainbowRodTrailErosion = 196
    RainbowRodTrailShape = 197
    MasterTrophyHeads = 198
    WhipMarkers = 199
    IceWhipSlicerBody = 200
    FinalFractal = 201
    EmpressButterflyJar = 202
    SlimeQueenHookChain = 203
    QueenSlimeMount = 204
    QueenSlimeSaddle = 205
    QueenSlimeWingFront = 206
    QueenSlimeWingBack = 207
    QueenSlimeHookRope = 208
    EmpressBladeTrail = 209
    EmpressBladeErosion = 210
    BiomeTorchIconUI = 211
    PlaguebringerFlame = 212
    TimelessTravelerFlame = 213
    FloretProtectorFrontTube = 214
    CreditsRollNPC_DryadTurningToTree = 215
    CreditsRollScene_Mask = 216
    CreditsRollScene_DryadInForest = 217
    CreditsRollScene_GuideDoorClosed = 218
    CreditsRollScene_GuideDoorOpened = 219
    CreditsRollScene_SkeletonMerchantCave = 220
    CreditsRollScene_Hallow = 221
    CreditsRollScene_CampfireBeach = 222
    CreditsRollScene_MerchantsGuild = 223
    CreditsRollScene_GolfCourse = 224
    CreditsRollScene_TruffleHouse = 225
    CreditsRollObject_Campfire = 226
    CreditsRollObject_BonePile = 227
    CreditsRollNPC_SkeletonMerchantScavenging = 228
    CreditsRollNPC_CyborgUnderRepair = 229
    CreditsRollNPC_SteampunkerRepairing = 230
    CreditsRollScene_Jungle = 231
    CreditsRollScene_SteampunkerWorkshop = 232
    CreditsRollScene_SteampunkerWorkshop2 = 233
    CreditsRollScene_ForestBar = 234
    CreditsRollScene_TavernkeepTavern = 235
    CreditsRollScene_ChristmasTree = 236
    CreditsRollScene_TinkererWorkshop = 237
    CreditsRollObject_Balloons = 238
    CreditsRollObject_Fireplace = 239
    CreditsRollObject_MiniPortal = 240
    CreditsRollObject_FinaleScene = 241
    PotionOfReturnGateInOutline = 242
    WorkshopIcon = 243
    ChesterOutline = 244
    DeerclopsEye = 245
    FartMinecart = 246
    StardewValleyPortal = 247
    StardewValleyPortalMask = 248
    StardewValleyPortalAnimation = 249
    TownSlimeRainbowOverlay = 250
    TerraFartMinecart = 251
    LightDisc = 252
    WolfMountBody = 253
    WolfMountHead = 254
    SuperCartEnabled = 255
    SuperCartDisabled = 256
    RubbleMakerIndicator = 257
    UnsafeIndicator = 258
    Loadouts = 259
    ShimmerflyinaBottle = 260
    JimsDroneRadio = 261
    RoamingFly = 262
    ShimmeredMechanicWrench = 263
    GuideShimmerFrontal = 264
    TravellingMerchantShimmerFrontal = 265
    DeadCellsBeheadedHead = 266
    GenericCircle = 267
    ShockIcon = 268
    DeadCellsBarrelExplosion = 269
    MoonlordWhipEye = 270
    VelociraptorMountBody = 271
    VelociraptorMountArm = 272
    RatMountBody = 273
    BatMountBody = 274
    JimCloakShoulder = 275
    JimsHelmetMetal = 276
    JimsBreastplateMetal = 277
    JimsLeggingsMetal = 278
    ChippysHeadbandTassels = 279
    SunSunglasses = 280
    FakeButterfly = 281
    TeamBasedSpawnMapIcons = 282
    RainbowBoulderPetAura = 283
    HeroicisHeadMount = 284
    AcornSlingshotHeldItem = 285
    AuroraNoise = 286
    AuroraShape = 287
    LightningShape = 288
    LibrarianSkeletonCircle = 289
    PixieMountBackLayer = 290
    PixieMountFrontLayer = 291
    HouseCheckBlock = 292
    HouseCheckDot = 293
    HouseCheckMarkerLight = 294
    HouseCheckMarkerChair = 295
    HouseCheckMarkerDoor = 296
    HouseCheckMarkerTable = 297
    HouseCheckHammer = 298
    PylonOffscreenIndicator = 299
    LunaHeadEars = 300
    Chillet = 301
    ChilletIgnis = 302
    FoxsparksHeld = 303




from dataclasses import dataclass

from terrex.net.drawing.particle_orchestra_settings import ParticleOrchestraSettings
from terrex.net.streamer import Reader, Writer

from .net_module import NetSyncModule


@dataclass()
class NetParticlesModule(NetSyncModule):
    id: int = 8
    particle_orchestra_type: int = 0
    particle_orchestra_settings: ParticleOrchestraSettings | None = None

    @classmethod
    def create(
        cls, particle_orchestra_type: int, particle_orchestra_settings: ParticleOrchestraSettings
    ) -> "NetParticlesModule":
        obj = cls()
        obj.particle_orchestra_type = particle_orchestra_type
        obj.particle_orchestra_settings = particle_orchestra_settings
        return obj

    def read(self, reader: Reader) -> None:
        self.particle_orchestra_type = reader.read_byte()
        self.particle_orchestra_settings = ParticleOrchestraSettings.read(reader)

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.particle_orchestra_type)
        if self.particle_orchestra_settings is None:
            raise ValueError("particle_orchestra_settings must not be None")
        self.particle_orchestra_settings.write(writer)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class AchievementMessageEventHappened(ServerPacket):
    id = MessageID.AchievementMessageEventHappened

    def __init__(self, event_id: int = 0):
        self.event_id = event_id

    def read(self, reader: Reader) -> None:
        self.event_id = reader.read_short()



from dataclasses import dataclass

from terrex.net.structure.rgb import Rgb as Color
from terrex.world.world_gen import WorldGen


@dataclass
class MapTile:
    type: int
    color: int
    light: int

    def __str__(self):
        return f"MapTile(type={self.type}, color={self.color}, light={self.light})"


class MapHelper:
    # tile_colors: Dict[int, List[Color]] = {}
    # wall_colors: Dict[int, List[Color]] = {}
    color_lookup: list[Color] = []
    tile_option_counts: list[int] = []
    wall_option_counts: list[int] = []
    tile_lookup: list[int] = []
    wall_lookup: list[int] = []
    tile_position: int = 0
    wall_position: int = 0
    wall_range_start: int = 0
    wall_range_end: int = 0
    liquid_position: int = 0
    sky_position: int = 0
    dirt_position: int = 0
    rock_position: int = 0
    hell_position: int = 0
    snow_types: list[int] = []

    @classmethod
    def initialize(cls):
        from terrex.world.map_colors import add_colors

        add_colors()

    @classmethod
    def map_color(cls, tile_type: int, old_color: Color, color_type: int) -> Color:
        paint = WorldGen.paint_color(color_type)

        r = old_color.r / 255.0
        g = old_color.g / 255.0
        b = old_color.b / 255.0

        max_ch = r
        if g > max_ch:
            max_ch = g
        if b > max_ch:
            max_ch = b

        if color_type == 29:
            factor = b * 0.3
            new_color = Color(int(paint.r * factor), int(paint.g * factor), int(paint.b * factor))
        elif color_type == 30:
            # invert
            new_r = 255 - old_color.r
            new_g = 255 - old_color.g
            new_b = 255 - old_color.b
            if cls.wall_range_start <= tile_type <= cls.wall_range_end:
                new_r = int(new_r * 0.5)
                new_g = int(new_g * 0.5)
                new_b = int(new_b * 0.5)
            new_color = Color(new_r, new_g, new_b)
        else:
            new_color = Color(int(paint.r * max_ch), int(paint.g * max_ch), int(paint.b * max_ch))
        return new_color

    @classmethod
    def get_map_tile_xna_color(cls, tile: MapTile) -> Color:
        color = cls.color_lookup[tile.type]
        color_type = tile.color
        if color_type > 0:
            color = cls.map_color(tile.type, color, color_type)
        if tile.light == 255:
            return color
        light = tile.light / 255.0
        return Color(int(color.r * light), int(color.g * light), int(color.b * light))

    @classmethod
    def create_map_tile(
        cls, world, x: int, y: int, base_light: int, background_override: int = 0
    ) -> MapTile:
        from terrex.net.structure.tile import Tile
        from terrex.world.world import World

        if not isinstance(world, World):
            raise TypeError("world must be a World instance")

        tile = world.tiles.get(x, y)
        if not isinstance(tile, Tile):
            return MapTile(0, 0, 0)

        color = 0
        light = base_light
        base_type = 0
        base_option = 0
        color, light, base_type, base_option = cls.get_tile_type(
            world, x, y, tile, color, light, base_type, base_option
        )
        if base_type == 0:
            color, light, base_type, base_option = cls.get_wall_type(
                x, y, tile, color, light, base_type, base_option
            )
        if base_type == 0:
            color = 0
            light = base_light
            if background_override == 0:
                base_type, light = cls.get_background_type(world, x, y, light)
            else:
                base_type = background_override
        return MapTile(base_type + base_option, color, light)

    @classmethod
    def get_tile_type(
        cls,
        world,
        x: int,
        y: int,
        tile,
        new_color: int,
        new_light: int,
        base_type: int,
        base_option: int,
    ) -> tuple[int, int, int, int]:
        from terrex.net.structure.tile import Tile
        from terrex.world.world import World

        if not isinstance(world, World):
            raise TypeError("world must be a World instance")

        if not isinstance(tile, Tile):
            raise TypeError("tile must be a Tile instance")

        debug_show_unbreakable = getattr(cls, 'show_unbreakable_wall', False)
        if (not debug_show_unbreakable or tile.wall != 350) and tile.active:
            tile_type = tile.type
            base_type = cls.tile_lookup[tile_type]
            is_invisible_block = tile.invisible_block
            if tile.fullbright_block and not is_invisible_block:
                new_light = 255
            if is_invisible_block:
                base_type = 0
                return new_color, new_light, base_type, base_option
            if tile_type == 5:
                # stub WorldGen.IsThisAMushroomTree(x, y)
                # if False:
                #     base_option = 1
                new_color = tile.color
                return new_color, new_light, base_type, base_option
            special_parity = False
            if tile_type <= 51:
                if tile_type == 19:
                    if tile.frame_y == 864:
                        base_type = 0
                elif tile_type == 51:
                    special_parity = True
            elif tile_type == 184:
                if tile.frame_x // 22 == 10:
                    tile_type = 627
                    base_type = cls.tile_lookup[tile_type]
            elif tile_type == 697:
                special_parity = True
            if special_parity:
                if (x + y) % 2 == 0:
                    base_type = 0
            if base_type != 0:
                base_option = cls.get_tile_base_option(world, x, y, tile_type, tile, base_option)
                if tile_type == 160:
                    new_color = 0
                    return new_color, new_light, base_type, base_option
                new_color = tile.color
        return new_color, new_light, base_type, base_option

    @classmethod
    def get_wall_type(
        cls, x: int, y: int, tile, new_color: int, new_light: int, base_type: int, base_option: int
    ) -> tuple[int, int, int, int]:
        from terrex.id import WallID
        from terrex.net.structure.tile import Tile

        if not isinstance(tile, Tile):
            raise TypeError("tile must be a Tile instance")

        is_invisible_wall = tile.invisible_wall
        if tile.wall > 0 and tile.fullbright_wall and not is_invisible_wall:
            new_light = 255
        if tile.liquid > 32:
            num = tile.liquid_type
            base_type = cls.liquid_position + num
            return new_color, new_light, base_type, base_option
        if not tile.invisible_wall and tile.wall > 0 and tile.wall < WallID.Count:
            num1 = tile.wall
            base_type = cls.wall_lookup[num1]
            new_color = tile.wall_color
            if num1 <= 27:
                if num1 == 21:
                    new_color = 0
                    return new_color, new_light, base_type, base_option
                if num1 == 27:
                    base_option = x % 2
                    return new_color, new_light, base_type, base_option
                base_option = 0
                return new_color, new_light, base_type, base_option
            elif num1 - 88 > 5 and num1 != 168 and num1 != 241:
                base_option = 0
                return new_color, new_light, base_type, base_option
            new_color = 0
            return new_color, new_light, base_type, base_option
        return new_color, new_light, base_type, base_option

    @classmethod
    def get_background_type(cls, world, x: int, y: int, light: int) -> tuple[int, int]:
        from terrex.world.world import World

        if not isinstance(world, World):
            raise TypeError("world must be a World instance")

        if y < world.world_surface:
            if world.remix_world:
                light = 5
                return 100, light
            light = 255
            bg_type = cls.calc_sky_gradient(world, cls.sky_position, 256, y)
            return bg_type, light
        if y >= world.underworld_layer():
            return cls.hell_position, light
        num = 0

        if world.generating_world is False:
            # stub sceneArea.Contains(x, y)
            if False:
                scene_snowiness = getattr(cls, 'scene_snowiness', 0.0)
                num = int(max(0.0, min(scene_snowiness, 1.0)) * 255)
            else:
                found = False
                for x1 in range(x - 36, x + 31, 10):
                    if found:
                        break
                    for y1 in range(y - 36, y + 31, 10):
                        # type = Main.Map[x1, y1].Type  # minimap type, stub
                        map_tile = world.tiles.get(x1, y1)
                        map_type = map_tile.type if map_tile is not None else 0
                        for snow_type in cls.snow_types:
                            if snow_type == map_type:
                                num = 255
                                found = True
                                break
                        if found:
                            break

        if y < world.rock_layer:
            return cls.dirt_position + num, light
        return cls.rock_position + num, light

    @classmethod
    def get_tile_base_option(
        cls, world, x: int, y: int, tile_type: int, tile, base_option: int
    ) -> int:
        from terrex.net.structure.tile import Tile
        from terrex.world.world import World

        if not isinstance(world, World):
            raise TypeError("world must be a World instance")

        if not isinstance(tile, Tile):
            raise TypeError("tile must be a Tile instance")

        if tile_type == 89:
            frame_div = tile.frame_x // 54
            if frame_div in (0, 21, 23):
                base_option = 0
            elif frame_div == 43:
                base_option = 2
            else:
                base_option = 1
        elif tile_type in (160, 627, 628, 692):
            base_option = (x + y) % 9
        elif tile_type == 461:
            if world.players[world.my_player_id].zone.corrupt:
                base_option = 1
            elif world.players[world.my_player_id].zone.crimson:
                base_option = 2
            elif world.players[world.my_player_id].zone.hallow:
                base_option = 3
            # if Main.scene_metrics.zone_corrupt:
            #     base_option = 1
            # elif Main.scene_metrics.zone_crimson:
            #     base_option = 2
            # elif Main.scene_metrics.zone_hallow:
            #     base_option = 3
        elif tile_type == 80:
            evil, good, crimson = WorldGen.get_cactus_type(world, x, y, tile.frame_x, tile.frame_y)
            if evil:
                base_option = 1
            elif good:
                base_option = 2
            elif crimson:
                base_option = 3
            else:
                base_option = 0
        elif tile_type == 529:
            num9 = y + 1
            corrupt_count2, crimson_count2, hallowed_count2 = WorldGen.get_biome_influence(
                world, x, x, num9, num9
            )
            num10 = max(corrupt_count2, crimson_count2, hallowed_count2)
            if corrupt_count2 == 0 and crimson_count2 == 0 and hallowed_count2 == 0:
                base_option = (
                    1
                    if x < WorldGen.beach_distance
                    or x > world.max_tiles_x - WorldGen.beach_distance
                    else 0
                )
            elif hallowed_count2 == num10:
                base_option = 2
            elif crimson_count2 != num10:
                base_option = 4
            else:
                base_option = 3
        elif tile_type == 530:
            num2 = y - (tile.frame_y % 36) // 18 + 2
            num3 = x - (tile.frame_x % 54) // 18
            corrupt_count, crimson_count, hallowed_count = WorldGen.get_biome_influence(
                world, num3, num3 + 3, num2, num2
            )
            num4 = max(corrupt_count, crimson_count, hallowed_count)
            if corrupt_count != 0 or crimson_count != 0 or hallowed_count != 0:
                if hallowed_count == num4:
                    base_option = 1
                elif crimson_count != num4:
                    base_option = 3
                else:
                    base_option = 2
            else:
                base_option = 0
        elif tile_type == 19:
            num13 = tile.frame_y // 18
            base_option = 1 if num13 == 48 else 0
        elif tile_type == 15:
            num8 = tile.frame_y // 40
            base_option = 1 if num8 in (1, 20) else 0
        elif tile_type in (518, 519):
            base_option = tile.frame_y // 18
        elif tile_type == 4:
            base_option = 1 if tile.frame_x < 66 else 0
        elif tile_type == 572:
            base_option = tile.frame_y // 36
        elif tile_type in (21, 441):
            frame_div = tile.frame_x // 36
            if frame_div in (1, 2, 10, 13, 15):
                base_option = 1
            elif frame_div in (3, 4):
                base_option = 2
            elif frame_div == 6:
                base_option = 3
            elif frame_div in (11, 17):
                base_option = 4
            else:
                base_option = 0
        elif tile_type in (467, 468):
            num = tile.frame_x // 36
            if num in range(0, 12):
                base_option = num
            elif num in (12, 13, 15, 18):
                base_option = 10
            elif num in (14, 17):
                base_option = 7
            elif num == 16:
                base_option = 3
            else:
                base_option = 12
        elif tile_type == 560:
            num = tile.frame_x // 36
            base_option = num if 0 <= num <= 2 else 0
        elif tile_type in (28, 653):
            fy = tile.frame_y
            if fy < 144:
                base_option = 0
            elif fy < 252:
                base_option = 1
            elif fy < 360 or (900 <= fy < 1008):
                base_option = 2
            elif fy < 468:
                base_option = 3
            elif fy < 576:
                base_option = 4
            elif fy < 684:
                base_option = 5
            elif fy < 792:
                base_option = 6
            elif fy < 900:
                base_option = 8
            elif fy < 1008:
                base_option = 7
            elif fy < 1116:
                base_option = 0
            elif fy < 1224:
                base_option = 3
            else:
                base_option = 7
        elif tile_type == 27:
            base_option = 1 if tile.frame_y < 34 else 0
        elif tile_type in (31, 696):
            base_option = 1 if tile.frame_x >= 36 else 0
        elif tile_type in (26, 695):
            base_option = 1 if tile.frame_x >= 54 else 0
        elif tile_type == 137:
            fy_div = tile.frame_y // 18
            if fy_div in (1, 2, 3, 4):
                base_option = 1
            elif fy_div == 5:
                base_option = 2
            else:
                base_option = 0
        elif tile_type in (82, 83, 84):
            fx = tile.frame_x
            if fx < 18:
                base_option = 0
            elif fx < 36:
                base_option = 1
            elif fx < 54:
                base_option = 2
            elif fx < 72:
                base_option = 3
            elif fx < 90:
                base_option = 4
            elif fx < 108:
                base_option = 5
            else:
                base_option = 6
        elif tile_type == 591:
            base_option = tile.frame_x // 36
        elif tile_type == 105:
            fx = tile.frame_x
            if 1548 <= fx <= 1654:
                base_option = 1
            elif 1656 <= fx <= 1798:
                base_option = 2
            else:
                base_option = 0
        elif tile_type == 133:
            base_option = 0 if tile.frame_x < 52 else 1
        elif tile_type == 134:
            base_option = 0 if tile.frame_x < 28 else 1
        elif tile_type == 149:
            base_option = y % 3
        elif tile_type in (165, 693, 694):
            fx = tile.frame_x
            if fx < 54:
                base_option = 0
            elif fx < 106:
                base_option = 1
            elif fx >= 216:
                base_option = 1
            elif fx < 162:
                base_option = 2
            else:
                base_option = 3
        elif tile_type == 178:
            fx = tile.frame_x
            if fx < 18:
                base_option = 0
            elif fx < 36:
                base_option = 1
            elif fx < 54:
                base_option = 2
            elif fx < 72:
                base_option = 3
            elif fx < 90:
                base_option = 4
            elif fx < 108:
                base_option = 5
            else:
                base_option = 6
        elif tile_type == 184:
            fx = tile.frame_x
            if fx < 22:
                base_option = 0
            elif fx < 44:
                base_option = 1
            elif fx < 66:
                base_option = 2
            elif fx < 88:
                base_option = 3
            elif fx < 110:
                base_option = 4
            elif fx < 132:
                base_option = 5
            elif fx < 154:
                base_option = 6
            elif fx < 176:
                base_option = 7
            elif fx < 198:
                base_option = 8
            elif fx < 220:
                base_option = 9
            elif fx < 242:
                base_option = 10
        elif tile_type == 650:
            num = tile.frame_x // 18
            if num < 6 or num in (28, 29, 30, 31, 32):
                base_option = 0
            elif num < 12 or num in (33, 34, 35):
                base_option = 1
            elif num < 28:
                base_option = 2
            elif num < 48:
                base_option = 3
            elif num < 54:
                base_option = 4
            elif num < 72:
                base_option = 0
            elif num == 72:
                base_option = 1
            elif num < 78:
                base_option = 11
        elif tile_type == 649:
            num = tile.frame_x // 36
            num6 = tile.frame_y // 18 - 1
            num += num6 * 18
            if num < 6 or num in (19, 20, 21, 22, 23, 24, 33, 38, 39, 40):
                base_option = 0
            elif num < 16:
                base_option = 2
            elif num < 19 or num in (31, 32):
                base_option = 1
            elif num < 31:
                base_option = 3
            elif num < 38:
                base_option = 4
            elif num < 59:
                base_option = 0
            elif num < 62:
                base_option = 1
            elif num < 65:
                base_option = 11
        elif tile_type == 185:
            if tile.frame_y < 18:
                num = tile.frame_x // 18
                if num < 6 or num in (28, 29, 30, 31, 32):
                    base_option = 0
                elif num < 12 or num in (33, 34, 35):
                    base_option = 1
                elif num < 28:
                    base_option = 2
                elif num < 48:
                    base_option = 3
                elif num < 54:
                    base_option = 4
                elif num < 72:
                    base_option = 0
                elif num == 72:
                    base_option = 1
                elif num < 78:
                    base_option = 11
            else:
                num = tile.frame_x // 36
                num12 = tile.frame_y // 18 - 1
                num += num12 * 18
                if num < 6 or num in (19, 20, 21, 22, 23, 24, 33, 38, 39, 40):
                    base_option = 0
                elif num < 16:
                    base_option = 2
                elif num < 19 or num in (31, 32):
                    base_option = 1
                elif num < 31:
                    base_option = 3
                elif num < 38:
                    base_option = 4
                elif num < 59:
                    base_option = 0
                elif num < 62:
                    base_option = 1
                elif num < 65:
                    base_option = 11
        elif tile_type in (186, 647):
            num = tile.frame_x // 54
            if num < 7:
                base_option = 2
            elif num < 22 or num in (33, 34, 35):
                base_option = 0
            elif num < 25:
                base_option = 1
            elif num == 25:
                base_option = 5
            elif num < 32:
                base_option = 3
        elif tile_type in (187, 648):
            num = tile.frame_x // 54
            num7 = tile.frame_y // 36
            num += num7 * 36
            if num < 3 or num in (14, 15, 16):
                base_option = 0
            elif num < 6:
                base_option = 6
            elif num < 9:
                base_option = 7
            elif num < 14:
                base_option = 4
            elif num < 18:
                base_option = 4
            elif num < 23:
                base_option = 8
            elif num < 25:
                base_option = 0
            elif num < 29:
                base_option = 1
            elif num < 47:
                base_option = 0
            elif num < 50:
                base_option = 1
            elif num < 52:
                base_option = 10
            elif num < 55:
                base_option = 2
        elif tile_type == 227:
            base_option = tile.frame_x // 34
        elif tile_type == 129:
            base_option = 1 if tile.frame_x >= 324 else 0
        elif tile_type == 240:
            num = tile.frame_x // 54
            num14 = tile.frame_y // 54
            num += num14 * 36
            if (0 <= num <= 11) or (47 <= num <= 53) or num in (72, 73, 75):
                base_option = 0
            elif (num < 12 or num > 15) and (num < 18 or num > 35) and (num < 63 or num > 71):
                if num in (16, 17):
                    base_option = 2
                elif num in (41, 42, 43, 44, 45):
                    base_option = 3
                elif num == 46:
                    base_option = 4
                else:
                    base_option = 1
            else:
                base_option = 1
        elif tile_type == 242:
            num = tile.frame_y // 72
            base_option = 1 if tile.frame_x // 106 == 0 and 22 <= num <= 24 else 0
        elif tile_type == 440:
            num = tile.frame_x // 54
            base_option = min(num, 6)
        elif tile_type == 457:
            num = tile.frame_x // 36
            base_option = min(num, 4)
        elif tile_type == 453:
            num = tile.frame_x // 36
            base_option = min(num, 2)
        elif tile_type == 419:
            num = tile.frame_x // 18
            base_option = min(num, 2)
        elif tile_type == 428:
            num = tile.frame_y // 18
            base_option = min(num, 3)
        elif tile_type == 420:
            num = tile.frame_y // 18
            base_option = min(num, 5)
        elif tile_type == 423:
            num = tile.frame_y // 18
            base_option = min(num, 6)
        elif tile_type == 493:
            fx = tile.frame_x
            if fx < 18:
                base_option = 0
            elif fx < 36:
                base_option = 1
            elif fx < 54:
                base_option = 2
            elif fx < 72:
                base_option = 3
            elif fx < 90:
                base_option = 4
            else:
                base_option = 5
        elif tile_type == 548:
            base_option = 0 if tile.frame_x // 54 < 7 else 1
        elif tile_type == 597:
            num = tile.frame_x // 54
            base_option = num if num <= 10 else 0
        else:
            base_option = 0

        return base_option

    @classmethod
    def calc_sky_gradient(cls, world, sky_position: int, max_sky_gradients: int, y: int) -> int:
        from terrex.world.world import World

        if not isinstance(world, World):
            raise TypeError("world must be a World instance")

        world_surface = world.world_surface
        num = int((max_sky_gradients - 1) * (y / world_surface))
        num = min(255, num)  # byte
        return sky_position + num


MapHelper.initialize()



from terrex.id import MessageID
from terrex.net.streamer import Writer
from terrex.packet.base import ClientPacket


class BugCatching(ClientPacket):
    id = MessageID.BugCatching

    def __init__(self, npc_id: int = 0, player_id: int = 0):
        self.npc_id = npc_id
        self.player_id = player_id

    def write(self, writer: Writer):
        writer.write_short(self.npc_id)
        writer.write_byte(self.player_id)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class DamageNPC(SyncPacket):
    id = MessageID.DamageNPC

    def __init__(
        self,
        npc_id: int = 0,
        damage: int = 0,
        knockback: float = 0.0,
        hit_direction: int = 0,
        crit: bool = False,
    ):
        self.npc_id = npc_id
        self.damage = damage
        self.knockback = knockback
        self.hit_direction = hit_direction
        self.crit = crit

    def read(self, reader: Reader):
        self.npc_id = reader.read_short()
        self.damage = reader.read_short()
        self.knockback = reader.read_float()
        self.hit_direction = reader.read_byte()
        self.crit = reader.read_bool()

    def write(self, writer: Writer):
        writer.write_short(self.npc_id)
        writer.write_short(self.damage)
        writer.write_float(self.knockback)
        writer.write_byte(self.hit_direction)
        writer.write_bool(self.crit)



"""
ItemSourceID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class ItemSourceID(IntEnum):
    NoneValue = 0
    SetBonus_Nebula = 1
    LuckyCoin = 2
    PlayerDeath = 3
    PlayerDrop = 4
    GrandDesignOrMultiColorWrench = 5
    TorchGod = 6
    Shimmer = 8
    Digesting = 9
    InventoryOverflow = 10




from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer

from .base import ClientPacket


class SpawnBossUseLicenseStartEvent(ClientPacket):
    id = MessageID.SpawnBossUseLicenseStartEvent

    def __init__(self) -> None:
        self.player_id: int = 0
        self.invasion_type: int = 0  # positive for bosses, negative for invasions

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_short()
        self.invasion_type = reader.read_short()

    def write(self, writer: Writer) -> None:
        writer.write_short(self.player_id)
        writer.write_short(self.invasion_type)



import inspect
from abc import ABC, abstractmethod
from typing import Any

from terrex.net.streamer import Reader, Writer
from terrex.util.stringify import stringify_value

net_module_registry: dict[int, type["NetModule"]] = {}


class NetModule(ABC):
    id: int

    def __init_subclass__(cls):
        super().__init_subclass__()

        if inspect.isabstract(cls):
            return

        if not hasattr(cls, "id"):
            raise TypeError(f"{cls.__name__} must define class attribute 'id'")

        if "create" not in cls.__dict__:
            raise TypeError(f"{cls.__name__} must implement classmethod create()")

        if cls.id in net_module_registry:
            raise ValueError(
                f"Module id {cls.id} already registered for {net_module_registry[cls.id]}"
            )
        net_module_registry[cls.id] = cls

    def __to_log_dict(self) -> dict[str, Any]:
        return {name: stringify_value(value) for name, value in vars(self).items()}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__to_log_dict()})"

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs):
        raise NotImplementedError(f"create() not implemented in {cls.__class__.__name__}")

    @abstractmethod
    def read(self, reader: Reader) -> None:
        raise NotImplementedError(f"read() not implemented in {self.__class__.__name__}")

    @abstractmethod
    def write(self, writer: Writer) -> None:
        raise NotImplementedError(f"_write() not implemented in {self.__class__.__name__}")


class NetServerModule(NetModule, ABC):
    """
    Server -> Client
    """


class NetClientModule(NetModule, ABC):
    """
    Client -> Server
    """


class NetSyncModule(NetModule, ABC):
    """
    Server <-> Client (Sync)
    """



from dataclasses import dataclass
from enum import IntEnum

from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


class TagEffectMessageType(IntEnum):
    FullState = 0
    ChangeActiveEffect = 1
    ApplyTagToNPC = 2
    EnableProcOnNPC = 3
    ClearProcOnNPC = 4


@dataclass()
class NetTagEffectModule(NetServerModule):
    id: int = 12
    player_id: int | None = None
    msg_type: TagEffectMessageType | None = None
    effect_id: int | None = None
    npc_index: int | None = None
    time_left_sparse: list[tuple[int, int]] | None = None
    proc_time_sparse: list[tuple[int, int]] | None = None

    @classmethod
    def create(
        cls,
        player_id: int,
        msg_type: TagEffectMessageType,
        effect_id: int | None = None,
        npc_index: int | None = None,
        time_left_sparse: list[tuple[int, int]] | None = None,
        proc_time_sparse: list[tuple[int, int]] | None = None,
    ) -> "NetTagEffectModule":
        obj = cls()
        obj.player_id = player_id
        obj.msg_type = msg_type
        obj.effect_id = effect_id
        obj.npc_index = npc_index
        obj.time_left_sparse = time_left_sparse
        obj.proc_time_sparse = proc_time_sparse
        return obj

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.msg_type = TagEffectMessageType(reader.read_byte())
        self.effect_id = None
        self.npc_index = None
        self.time_left_sparse = None
        self.proc_time_sparse = None
        if self.msg_type == TagEffectMessageType.FullState:
            self.effect_id = reader.read_short()
            self.time_left_sparse = self._read_sparse(reader)
            self.proc_time_sparse = self._read_sparse(reader)
        elif self.msg_type == TagEffectMessageType.ChangeActiveEffect:
            self.effect_id = reader.read_short()
        elif self.msg_type in (
            TagEffectMessageType.ApplyTagToNPC,
            TagEffectMessageType.EnableProcOnNPC,
            TagEffectMessageType.ClearProcOnNPC,
        ):
            self.npc_index = reader.read_byte()

    @classmethod
    def _read_sparse(cls, reader: Reader) -> list[tuple[int, int]]:
        sparse = []
        while True:
            idx = reader.read_byte()
            if idx >= 255:
                break
            time = reader.read_int()
            sparse.append((idx, time))
        return sparse

    def write(self, writer: Writer) -> None:
        if self.player_id is None or self.msg_type is None:
            raise ValueError("player_id and msg_type must be int/TagEffectMessageType and not None")
        writer.write_byte(self.player_id)
        writer.write_byte(self.msg_type.value)
        if self.msg_type == TagEffectMessageType.FullState:
            if self.effect_id is None:
                raise ValueError("effect_id must not be None for FullState")
            writer.write_short(self.effect_id)
            self._write_sparse(writer, self.time_left_sparse or [])
            self._write_sparse(writer, self.proc_time_sparse or [])
        elif self.msg_type == TagEffectMessageType.ChangeActiveEffect:
            if self.effect_id is None:
                raise ValueError("effect_id must not be None for ChangeActiveEffect")
            writer.write_short(self.effect_id)
        elif self.msg_type in (
            TagEffectMessageType.ApplyTagToNPC,
            TagEffectMessageType.EnableProcOnNPC,
            TagEffectMessageType.ClearProcOnNPC,
        ):
            if self.npc_index is None:
                raise ValueError("npc_index must not be None for these types")
            writer.write_byte(self.npc_index)

    def _write_sparse(self, writer: Writer, sparse: list[tuple[int, int]]) -> None:
        for idx, time in sparse:
            writer.write_byte(idx)
            writer.write_int(time)
        writer.write_byte(255)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class CrystalInvasionSendWaitTime(ServerPacket):
    id = MessageID.CrystalInvasionSendWaitTime

    def __init__(self, time_until_next_wave: int = 0):
        self.time_until_next_wave = time_until_next_wave

    def read(self, reader: Reader):
        self.time_until_next_wave = reader.read_int()



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.net.structure.player_death_reason import PlayerDeathReason
from terrex.packet.base import ClientPacket


class PlayerHurtV2(ClientPacket):
    id = MessageID.PlayerHurtV2

    def __init__(
        self,
        player_id: int = 0,
        reason: PlayerDeathReason | None = None,
        damage: int = 0,
        hit_direction: int = 0,
        flags: int = 0,
        cooldown_counter: int = 0,
    ):
        self.player_id = player_id
        self.reason = reason or PlayerDeathReason()
        self.damage = damage
        self.hit_direction = hit_direction
        self.flags = flags
        self.cooldown_counter = cooldown_counter

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.reason = PlayerDeathReason.read(reader)
        self.damage = reader.read_short()
        self.hit_direction = reader.read_byte()
        self.flags = reader.read_byte()
        self.cooldown_counter = reader.read_sbyte()

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.player_id)
        self.reason.write(writer)
        writer.write_short(self.damage)
        writer.write_byte(self.hit_direction)
        writer.write_byte(self.flags)
        writer.write_sbyte(self.cooldown_counter)
